"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
import pytest
from datetime import datetime, date, timedelta
from sqlalchemy.orm.session import Session
from sqlalchemy import asc
from decimal import Decimal

from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.scheduled_transaction import ScheduledTransaction
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.budget_transaction import BudgetTransaction
from biweeklybudget.tests.unit_helpers import binexp_to_dict
from biweeklybudget.utils import dtnow

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import Mock, patch, call, DEFAULT
else:
    from unittest.mock import Mock, patch, call, DEFAULT

pbm = 'biweeklybudget.biweeklypayperiod'
pb = '%s.BiweeklyPayPeriod' % pbm


class TestInit(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_init(self):
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        assert cls._start_date == date(2017, 3, 17)
        assert cls._end_date == date(2017, 3, 30)
        assert cls._db == self.mock_sess
        assert cls._data_cache == {}

    def test_init_datetime(self):
        cls = BiweeklyPayPeriod(datetime(2017, 3, 17), self.mock_sess)
        assert cls._start_date == date(2017, 3, 17)
        assert cls._end_date == date(2017, 3, 30)
        assert cls._db == self.mock_sess
        assert cls._data_cache == {}


class TestProperties(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_period_interval(self):
        assert self.cls.period_interval == timedelta(days=14)

    def test_period_length(self):
        assert self.cls.period_length == timedelta(days=13)

    def test_start_date(self):
        assert self.cls.start_date == date(2017, 3, 17)

    def test_end_date(self):
        assert self.cls.end_date == date(2017, 3, 30)

    def test_next(self):
        assert self.cls.next == BiweeklyPayPeriod(
            date(2017, 3, 31), self.mock_sess)

    def test_previous(self):
        assert self.cls.previous == BiweeklyPayPeriod(
            date(2017, 3, 3), self.mock_sess)


class TestMagicMethods(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_repr(self):
        assert str(self.cls) == '<BiweeklyPayPeriod(2017-03-17)>'

    @pytest.mark.skipif(sys.version_info[0] >= 3, reason='py2 only')
    def test_ordering_py27(self):
        assert self.cls < BiweeklyPayPeriod(date(2017, 4, 16), self.mock_sess)
        assert self.cls > BiweeklyPayPeriod(date(2017, 2, 13), self.mock_sess)
        assert self.cls == BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        self.cls < 2
        self.cls.__eq__(2)

    @pytest.mark.skipif(sys.version_info[0] < 3, reason='py3 only')
    def test_ordering_py3(self):
        assert self.cls < BiweeklyPayPeriod(date(2017, 4, 16), self.mock_sess)
        assert self.cls > BiweeklyPayPeriod(date(2017, 2, 13), self.mock_sess)
        assert self.cls == BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        with pytest.raises(TypeError):
            self.cls < 2
        self.cls.__eq__(2)


class TestForDate(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 3, 17))
    def test_period_for_date(self):
        assert BiweeklyPayPeriod.period_for_date(
            date(2017, 3, 16), self.mock_sess) == BiweeklyPayPeriod(
            date(2017, 3, 3), self.mock_sess)
        assert BiweeklyPayPeriod.period_for_date(
            date(2017, 3, 17), self.mock_sess) == BiweeklyPayPeriod(
            date(2017, 3, 17), self.mock_sess)
        assert BiweeklyPayPeriod.period_for_date(
            date(2017, 3, 18), self.mock_sess) == BiweeklyPayPeriod(
            date(2017, 3, 17), self.mock_sess)
        assert BiweeklyPayPeriod.period_for_date(
            date(2017, 3, 30), self.mock_sess) == BiweeklyPayPeriod(
            date(2017, 3, 17), self.mock_sess)
        assert BiweeklyPayPeriod.period_for_date(
            date(2017, 3, 31), self.mock_sess) == BiweeklyPayPeriod(
            date(2017, 3, 31), self.mock_sess)
        assert BiweeklyPayPeriod.period_for_date(
            date(2017, 1, 21), self.mock_sess) == BiweeklyPayPeriod(
            date(2017, 1, 20), self.mock_sess)
        assert BiweeklyPayPeriod.period_for_date(
            date(2017, 5, 2), self.mock_sess) == BiweeklyPayPeriod(
            date(2017, 4, 28), self.mock_sess)


class TestFilterQuery(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_filter_query(self):
        q = Mock()
        res = self.cls.filter_query(q, OFXTransaction.date_posted)
        assert len(q.mock_calls) == 1
        kall = q.mock_calls[0]
        assert kall[0] == 'filter'
        a = OFXTransaction.date_posted >= self.cls.start_date
        b = OFXTransaction.date_posted <= self.cls.end_date
        assert kall[1][0].compare(a) is True
        assert kall[1][1].compare(b) is True
        assert res == q.filter.return_value


class TestTransactions(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_transactions(self):
        mock_res = Mock()
        with patch('%s.filter_query' % pb, autospec=True) as mock_filter:
            mock_filter.return_value = mock_res
            res = self.cls._transactions()
        assert res == mock_res
        assert mock_filter.mock_calls == [
            call(self.cls, self.mock_sess.query.return_value, Transaction.date)
        ]
        assert self.mock_sess.mock_calls == [
            call.query(Transaction)
        ]


class TestSTDate(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_scheduled_transactions_date(self):
        mock_res = Mock()
        with patch('%s.filter_query' % pb, autospec=True) as mock_filter:
            mock_filter.return_value = mock_res
            res = self.cls._scheduled_transactions_date()
        assert res == mock_res
        assert mock_filter.mock_calls == [
            call(
                self.cls,
                self.mock_sess.query.return_value.filter.return_value,
                ScheduledTransaction.date
            )
        ]
        assert len(self.mock_sess.mock_calls) == 2
        assert self.mock_sess.mock_calls[0] == call.query(ScheduledTransaction)
        assert self.mock_sess.mock_calls[1][0] == 'query().filter'
        expected = ScheduledTransaction.is_active.__eq__(True)
        assert str(expected) == str(
            self.mock_sess.mock_calls[1][1][0]
        )


class TestSTPeriod(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_scheduled_transactions_per_period(self):
        res = self.cls._scheduled_transactions_per_period()
        frv = self.mock_sess.query.return_value.filter.return_value
        assert res == frv.order_by.return_value
        assert len(self.mock_sess.mock_calls) == 3
        assert self.mock_sess.mock_calls[0] == call.query(ScheduledTransaction)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected = ScheduledTransaction.schedule_type.__eq__('per period')
        assert binexp_to_dict(expected) == binexp_to_dict(kall[1][0])
        kall = self.mock_sess.mock_calls[2]
        assert kall[0] == 'query().filter().order_by'
        assert str(kall[1][0]) == str(asc(ScheduledTransaction.num_per_period))
        assert str(kall[1][1]) == str(asc(ScheduledTransaction.amount))


class TestSTMonthly(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)

    def test_contiguous(self):
        cls = BiweeklyPayPeriod(date(2017, 3, 2), self.mock_sess)
        res = cls._scheduled_transactions_monthly()
        assert res == self.mock_sess.query.return_value.filter.return_value
        assert len(self.mock_sess.mock_calls) == 2
        assert self.mock_sess.mock_calls[0] == call.query(ScheduledTransaction)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected = [
            ScheduledTransaction.schedule_type.__eq__('monthly'),
            ScheduledTransaction.is_active.__eq__(True),
            ScheduledTransaction.day_of_month.__le__(15),
            ScheduledTransaction.day_of_month.__ge__(2)
        ]
        assert binexp_to_dict(kall[1][0]) == binexp_to_dict(expected[0])
        assert str(kall[1][1]) == str(expected[1])
        assert binexp_to_dict(kall[1][2]) == binexp_to_dict(expected[2])
        assert binexp_to_dict(kall[1][3]) == binexp_to_dict(expected[3])

    def test_crossmonth(self):
        cls = BiweeklyPayPeriod(date(2017, 3, 24), self.mock_sess)
        mock_or_result = Mock()
        with patch('%s.or_' % pbm) as mock_or:
            mock_or.return_value = mock_or_result
            res = cls._scheduled_transactions_monthly()
        assert res == self.mock_sess.query.return_value.filter.return_value
        assert len(self.mock_sess.mock_calls) == 2
        assert self.mock_sess.mock_calls[0] == call.query(ScheduledTransaction)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected = ScheduledTransaction.schedule_type.__eq__('monthly')
        assert binexp_to_dict(kall[1][0]) == binexp_to_dict(expected)
        assert str(kall[1][1]) == str(
            ScheduledTransaction.is_active.__eq__(True)
        )
        assert kall[1][2] == mock_or_result
        kall = mock_or.mock_calls[0]
        assert len(mock_or.mock_calls) == 1
        expected = [
            ScheduledTransaction.day_of_month.__le__(6),
            ScheduledTransaction.day_of_month.__ge__(24)
        ]
        for idx, exp in enumerate(expected):
            assert binexp_to_dict(kall[1][idx]) == binexp_to_dict(exp)


class TestTransactionsList(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_simple(self):
        m = Mock()
        self.cls._data_cache = {'all_trans_list': m}
        assert self.cls.transactions_list == m


class TestBudgetSums(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_simple(self):
        m = Mock()
        self.cls._data_cache = {'budget_sums': m}
        assert self.cls.budget_sums == m


class TestOverallSums(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_simple(self):
        m = Mock()
        self.cls._data_cache = {'overall_sums': m}
        assert self.cls.overall_sums == m


class TestIncomeBudgetIDs(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 7), self.mock_sess)

    def test_no_cache(self):
        self.cls._income_budget_id_list = None
        budgets = [
            Mock(spec_set=Budget, is_income=True, id=2),
            Mock(spec_set=Budget, is_income=True, id=4)
        ]
        self.mock_sess.query.return_value.filter.return_value.all.return_value \
            = budgets
        assert self.cls._income_budget_ids == [2, 4]
        assert self.cls._income_budget_id_list == [2, 4]
        assert len(self.mock_sess.mock_calls) == 3
        assert self.mock_sess.mock_calls[0] == call.query(Budget)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        assert str(kall[1][0]) == str(Budget.is_income.__eq__(True))
        assert self.mock_sess.mock_calls[2] == call.query().filter().all()

    def test_cache(self):
        self.cls._income_budget_id_list = [1, 2]
        budgets = [
            Mock(spec_set=Budget, is_income=False, id=1),
            Mock(spec_set=Budget, is_income=True, id=2),
            Mock(spec_set=Budget, is_income=False, id=3),
            Mock(spec_set=Budget, is_income=True, id=4)
        ]
        self.mock_sess.query.return_value.filter.return_value.all.return_value \
            = budgets
        assert self.cls._income_budget_ids == [1, 2]
        assert self.cls._income_budget_id_list == [1, 2]
        assert len(self.mock_sess.mock_calls) == 0


class TestData(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_initial(self):
        mock_t = Mock()
        mock_std = Mock()
        mock_stpp = Mock()
        mock_stm = Mock()
        mock_stw = Mock()
        mock_sta = Mock()
        mock_sta_filtered = Mock()
        mock_mct = Mock()
        mock_mbs = Mock()
        mock_mos = Mock()
        with patch.multiple(
            pb,
            autospec=True,
            _transactions=DEFAULT,
            _scheduled_transactions_date=DEFAULT,
            _scheduled_transactions_per_period=DEFAULT,
            _scheduled_transactions_monthly=DEFAULT,
            _scheduled_transactions_weekly=DEFAULT,
            _scheduled_transactions_annual=DEFAULT,
            _filter_annual_for_period=DEFAULT,
            _make_combined_transactions=DEFAULT,
            _make_budget_sums=DEFAULT,
            _make_overall_sums=DEFAULT
        ) as mocks:
            mocks['_transactions'].return_value.all.return_value = mock_t
            mocks['_scheduled_transactions_date'
                  ''].return_value.all.return_value = mock_std
            mocks['_scheduled_transactions_per_period'
                  ''].return_value.all.return_value = mock_stpp
            mocks['_scheduled_transactions_monthly'
                  ''].return_value.all.return_value = mock_stm
            mocks['_scheduled_transactions_weekly'
                  ''].return_value.all.return_value = mock_stw
            mocks['_scheduled_transactions_annual'
                  ''].return_value.all.return_value = mock_sta
            mocks['_filter_annual_for_period'].return_value = mock_sta_filtered
            mocks['_make_combined_transactions'].return_value = mock_mct
            mocks['_make_budget_sums'].return_value = mock_mbs
            mocks['_make_overall_sums'].return_value = mock_mos
            res = self.cls._data
        assert res == {
            'transactions': mock_t,
            'st_date': mock_std,
            'st_per_period': mock_stpp,
            'st_monthly': mock_stm,
            'st_weekly': mock_stw,
            'st_annual': mock_sta_filtered,
            'all_trans_list': mock_mct,
            'budget_sums': mock_mbs,
            'overall_sums': mock_mos
        }
        assert mocks['_transactions'].mock_calls == [
            call(self.cls), call().all()
        ]
        assert mocks['_scheduled_transactions_date'].mock_calls == [
            call(self.cls), call().all()
        ]
        assert mocks['_scheduled_transactions_per_period'].mock_calls == [
            call(self.cls), call().all()
        ]
        assert mocks['_scheduled_transactions_monthly'].mock_calls == [
            call(self.cls), call().all()
        ]
        assert mocks['_scheduled_transactions_weekly'].mock_calls == [
            call(self.cls), call().all()
        ]
        assert mocks['_scheduled_transactions_annual'].mock_calls == [
            call(self.cls), call().all()
        ]
        assert mocks['_filter_annual_for_period'].mock_calls == [
            call(self.cls, mock_sta)
        ]
        assert mocks['_make_combined_transactions'].mock_calls == [
            call(self.cls)
        ]
        assert mocks['_make_budget_sums'].mock_calls == [
            call(self.cls)
        ]
        assert mocks['_make_overall_sums'].mock_calls == [
            call(self.cls)
        ]

    def test_cached(self):
        mock_t = Mock()
        mock_std = Mock()
        mock_stpp = Mock()
        mock_stm = Mock()
        mock_mct = Mock()
        mock_mbs = Mock()
        mock_mos = Mock()
        with patch.multiple(
            pb,
            autospec=True,
            _transactions=DEFAULT,
            _scheduled_transactions_date=DEFAULT,
            _scheduled_transactions_per_period=DEFAULT,
            _scheduled_transactions_monthly=DEFAULT,
            _make_combined_transactions=DEFAULT,
            _make_budget_sums=DEFAULT,
            _make_overall_sums=DEFAULT
        ) as mocks:
            mocks['_transactions'].return_value.all.return_value = mock_t
            mocks['_scheduled_transactions_date'
                  ''].return_value.all.return_value = mock_std
            mocks['_scheduled_transactions_per_period'
                  ''].return_value.all.return_value = mock_stpp
            mocks['_scheduled_transactions_monthly'
                  ''].return_value.all.return_value = mock_stm
            mocks['_make_combined_transactions'].return_value = mock_mct
            mocks['_make_budget_sums'].return_value = mock_mbs
            mocks['_make_overall_sums'].return_value = mock_mos
            self.cls._data_cache = {'foo': 'bar'}
            res = self.cls._data
        assert res == {'foo': 'bar'}
        assert mocks['_transactions'].mock_calls == []
        assert mocks['_scheduled_transactions_date'].mock_calls == []
        assert mocks['_scheduled_transactions_per_period'].mock_calls == []
        assert mocks['_scheduled_transactions_monthly'].mock_calls == []
        assert mocks['_make_combined_transactions'].mock_calls == []
        assert mocks['_make_budget_sums'].mock_calls == []
        assert mocks['_make_overall_sums'].mock_calls == []


class TestMakeCombinedTransactions(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_all(self):

        def se_trans_dict(_, t):
            if t._mock_name.startswith('per_period'):
                return {'name': t._mock_name, 'amount': t.amount}
            print(t)
            return {
                'date': date(year=2017, month=3, day=t.day),
                'name': t._mock_name,
                'amount': t.amount
            }

        mock_per_periodA = Mock(num_per_period=1, name='per_period_A', amount=8)
        mock_per_periodB = Mock(num_per_period=3, name='per_period_B', amount=9)
        self.cls._data_cache = {
            'transactions': [
                Mock(name='t1', day=1, scheduled_trans_id=1, amount=1),
                Mock(name='t2', day=4, scheduled_trans_id=None, amount=2),
                Mock(name='t3', day=6, scheduled_trans_id=2, amount=3)
            ],
            'st_date': [
                Mock(name='std1', day=2, id=1, amount=4),
                Mock(name='std2', day=5, amount=5)
            ],
            'st_monthly': [
                Mock(name='stm1', day=3, amount=6),
                Mock(name='stm2', day=6, id=2, amount=7)
            ],
            'st_per_period': [
                mock_per_periodA,
                mock_per_periodB
            ],
            'st_weekly': [],
            'st_annual': []
        }
        with patch('%s._trans_dict' % pb, autospec=True) as mock_t_dict:
            mock_t_dict.side_effect = se_trans_dict
            res = self.cls._make_combined_transactions()
        assert res == [
            {'name': 'per_period_A', 'amount': 8},
            {'name': 'per_period_B', 'amount': 9},
            {'name': 'per_period_B', 'amount': 9},
            {'name': 'per_period_B', 'amount': 9},
            {
                'name': 't1',
                'date': date(year=2017, month=3, day=1),
                'amount': 1
            },
            {
                'name': 'stm1',
                'date': date(year=2017, month=3, day=3),
                'amount': 6
            },
            {
                'name': 't2',
                'date': date(year=2017, month=3, day=4),
                'amount': 2
            },
            {
                'name': 'std2',
                'date': date(year=2017, month=3, day=5),
                'amount': 5
            },
            {
                'name': 't3',
                'date': date(2017, 3, 6),
                'amount': 3
            }
        ]


class TestMakeBudgetSums(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 7), self.mock_sess)

    def test_scheduled_income(self):
        self.cls._data_cache = {
            'all_trans_list': [
                {
                    'type': 'ScheduledTransaction',
                    'amount': Decimal('11.11'),
                    'budgeted_amount': None,
                    'budgets': {
                        1: {'name': 'foo', 'amount': Decimal('11.11')}
                    }
                },
                {
                    'type': 'Transaction',
                    'amount': Decimal('22.22'),
                    'budgeted_amount': None,
                    'budgets': {
                        1: {'name': 'foo', 'amount': Decimal('20.00')},
                        2: {'name': 'foo', 'amount': Decimal('2.22')}
                    }
                },
                {
                    'type': 'Transaction',
                    'amount': Decimal('22.22'),
                    'budgeted_amount': Decimal('20.20'),
                    'budgets': {
                        1: {'name': 'foo', 'amount': Decimal('22.22')}
                    },
                    'planned_budget_id': 1
                },
                {
                    'type': 'Transaction',
                    'amount': Decimal('33.33'),
                    'budgeted_amount': Decimal('33.33'),
                    'budgets': {
                        2: {'name': 'foo', 'amount': Decimal('33.33')}
                    },
                    'planned_budget_id': 2
                },
                {
                    'type': 'ScheduledTransaction',
                    'amount': Decimal('-1234.56'),
                    'budgeted_amount': Decimal('-1234.56'),
                    'budgets': {
                        4: {'name': 'foo', 'amount': Decimal('-1234.56')}
                    },
                    'planned_budget_id': 4
                }
            ]
        }
        budgets = [
            Mock(
                spec_set=Budget, starting_balance=Decimal('123.45'), id=1,
                is_income=False
            ),
            Mock(
                spec_set=Budget, starting_balance=Decimal('456.78'), id=2,
                is_income=False
            ),
            Mock(
                spec_set=Budget, starting_balance=Decimal('789.10'), id=3,
                is_income=False
            ),
            Mock(
                spec_set=Budget, starting_balance=Decimal('0'), id=4,
                is_income=True
            )
        ]
        self.mock_sess.query.return_value.filter.return_value.all.return_value \
            = budgets
        res = self.cls._make_budget_sums()
        assert res == {
            1: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('51.31'),
                'spent': Decimal('42.22'),
                'trans_total': Decimal('53.33'),
                'is_income': False,
                'remaining': Decimal('70.12')
            },
            2: {
                'budget_amount': Decimal('456.78'),
                'allocated': Decimal('35.55'),
                'spent': Decimal('35.55'),
                'trans_total': Decimal('35.55'),
                'is_income': False,
                'remaining': Decimal('421.23')
            },
            3: {
                'budget_amount': Decimal('789.10'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False,
                'remaining': Decimal('789.10')
            },
            4: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('-1234.56'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('-1234.56'),
                'is_income': True,
                'remaining': Decimal('1234.56')
            }
        }
        assert len(self.mock_sess.mock_calls) == 3
        assert self.mock_sess.mock_calls[0] == call.query(Budget)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected = [
            Budget.is_active.__eq__(True),
            Budget.is_periodic.__eq__(True)
        ]
        for idx, exp in enumerate(expected):
            assert str(kall[1][idx]) == str(expected[idx])
        assert self.mock_sess.mock_calls[2] == call.query().filter().all()

    def test_actual_income(self):
        self.cls._data_cache = {
            'all_trans_list': [
                {
                    'type': 'ScheduledTransaction',
                    'amount': Decimal('11.11'),
                    'budgeted_amount': None,
                    'budgets': {
                        1: {'name': 'foo', 'amount': Decimal('11.11')}
                    }
                },
                {
                    'type': 'Transaction',
                    'amount': Decimal('22.22'),
                    'budgeted_amount': None,
                    'budgets': {
                        1: {'name': 'foo', 'amount': Decimal('22.22')}
                    }
                },
                {
                    'type': 'Transaction',
                    'amount': Decimal('22.22'),
                    'budgeted_amount': Decimal('20.20'),
                    'budgets': {
                        1: {'name': 'foo', 'amount': Decimal('22.22')}
                    },
                    'planned_budget_id': 1
                },
                {
                    'type': 'Transaction',
                    'amount': Decimal('33.33'),
                    'budgeted_amount': Decimal('33.33'),
                    'budgets': {
                        2: {'name': 'foo', 'amount': Decimal('33.33')}
                    },
                    'planned_budget_id': 2
                },
                {
                    'type': 'Transaction',
                    'amount': Decimal('-1234.56'),
                    'budgeted_amount': Decimal('-1234.56'),
                    'budgets': {
                        4: {'name': 'foo', 'amount': Decimal('-1234.56')}
                    },
                    'planned_budget_id': 4
                }
            ]
        }
        budgets = [
            Mock(
                spec_set=Budget, starting_balance=Decimal('123.45'), id=1,
                is_income=False
            ),
            Mock(
                spec_set=Budget, starting_balance=Decimal('456.78'), id=2,
                is_income=False
            ),
            Mock(
                spec_set=Budget, starting_balance=Decimal('789.10'), id=3,
                is_income=False
            ),
            Mock(
                spec_set=Budget, starting_balance=Decimal('0'), id=4,
                is_income=True
            )
        ]
        self.mock_sess.query.return_value.filter.return_value.all.return_value \
            = budgets
        res = self.cls._make_budget_sums()
        assert res == {
            1: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('53.53'),
                'spent': Decimal('44.44'),
                'trans_total': Decimal('55.55'),
                'is_income': False,
                'remaining': Decimal('67.90')
            },
            2: {
                'budget_amount': Decimal('456.78'),
                'allocated': Decimal('33.33'),
                'spent': Decimal('33.33'),
                'trans_total': Decimal('33.33'),
                'is_income': False,
                'remaining': Decimal('423.45')
            },
            3: {
                'budget_amount': Decimal('789.10'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False,
                'remaining': Decimal('789.10')
            },
            4: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('-1234.56'),
                'spent': Decimal('-1234.56'),
                'trans_total': Decimal('-1234.56'),
                'is_income': True,
                'remaining': Decimal('1234.56')
            }
        }
        assert len(self.mock_sess.mock_calls) == 3
        assert self.mock_sess.mock_calls[0] == call.query(Budget)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected = [
            Budget.is_active.__eq__(True),
            Budget.is_periodic.__eq__(True)
        ]
        for idx, exp in enumerate(expected):
            assert str(kall[1][idx]) == str(expected[idx])
        assert self.mock_sess.mock_calls[2] == call.query().filter().all()


class TestMakeOverallSums(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)

    def test_past(self):
        cls = BiweeklyPayPeriod(date(2017, 3, 7), self.mock_sess)
        assert cls.is_in_past is True
        cls._data_cache['budget_sums'] = {
            1: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('53.53'),
                'spent': Decimal('44.44'),
                'trans_total': Decimal('55.55'),
                'is_income': False
            },
            2: {
                'budget_amount': Decimal('10.00'),
                'allocated': Decimal('33.33'),
                'spent': Decimal('33.33'),
                'trans_total': Decimal('33.33'),
                'is_income': False
            },
            3: {
                'budget_amount': Decimal('789.10'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False
            },
            4: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('-1192.56'),
                'spent': Decimal('-254.38'),
                'trans_total': Decimal('-1234.56'),
                'is_income': True
            }
        }
        assert cls._make_overall_sums() == {
            'allocated': Decimal('945.88'),
            'spent': Decimal('77.77'),
            'income': Decimal('1234.56'),
            'remaining': Decimal('1156.79')
        }

    def test_current(self):
        cls = BiweeklyPayPeriod(
            dtnow().date() - timedelta(days=4), self.mock_sess
        )
        assert cls.is_in_past is False
        cls._data_cache['budget_sums'] = {
            1: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('53.53'),
                'spent': Decimal('44.44'),
                'trans_total': Decimal('55.55'),
                'is_income': False
            },
            2: {
                'budget_amount': Decimal('10.00'),
                'allocated': Decimal('33.33'),
                'spent': Decimal('33.33'),
                'trans_total': Decimal('33.33'),
                'is_income': False
            },
            3: {
                'budget_amount': Decimal('789.10'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False
            },
            4: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('-1192.56'),
                'spent': Decimal('-254.38'),
                'trans_total': Decimal('-1234.56'),
                'is_income': True
            }
        }
        assert cls._make_overall_sums() == {
            'allocated': Decimal('945.88'),
            'spent': Decimal('77.77'),
            'income': Decimal('1234.56'),
            'remaining': Decimal('288.68')
        }

    def test_future(self):
        cls = BiweeklyPayPeriod(date(2017, 3, 7), self.mock_sess)
        assert cls.is_in_past is True
        cls._data_cache['budget_sums'] = {
            1: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('53.53'),
                'spent': Decimal('44.44'),
                'trans_total': Decimal('55.55'),
                'is_income': False
            },
            2: {
                'budget_amount': Decimal('10.00'),
                'allocated': Decimal('33.33'),
                'spent': Decimal('33.33'),
                'trans_total': Decimal('33.33'),
                'is_income': False
            },
            3: {
                'budget_amount': Decimal('789.10'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False
            },
            4: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('-1192.56'),
                'spent': Decimal('-254.38'),
                'trans_total': Decimal('-1234.56'),
                'is_income': True
            }
        }
        assert cls._make_overall_sums() == {
            'allocated': Decimal('945.88'),
            'spent': Decimal('77.77'),
            'income': Decimal('1234.56'),
            'remaining': Decimal('1156.79')
        }

    def test_issue201(self):
        cls = BiweeklyPayPeriod(
            dtnow().date() + timedelta(days=24), self.mock_sess
        )
        assert cls.is_in_past is False
        cls._data_cache['budget_sums'] = {
            2: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': True,
                'remaining': Decimal('123.45')
            },
            3: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('99.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('99.0'),
                'is_income': True,
                'remaining': Decimal('99.0')
            },
            4: {
                'budget_amount': Decimal('500.00'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False,
                'remaining': Decimal('500.0')
            },
            5: {
                'budget_amount': Decimal('100.0'),
                'allocated': Decimal('2.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('2.0'),
                'is_income': False,
                'remaining': Decimal('98.0')
            },
            6: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('2.0'),
                'spent': Decimal('200.0'),
                'trans_total': Decimal('200.0'),
                'is_income': False,
                'remaining': Decimal('-200.0')
            }
        }
        assert cls._make_overall_sums() == {
            'allocated': Decimal('602.0'),
            'income': Decimal('222.45'),
            'remaining': Decimal('-577.55'),
            'spent': Decimal('200.0')
        }


class TestTransDict(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_transaction(self):
        m = Mock(spec_set=Transaction)
        m_dst_res = Mock()
        m_dt_res = Mock()
        with patch('%s._dict_for_sched_trans' % pb, autospec=True) as m_dst:
            with patch('%s._dict_for_trans' % pb, autospec=True) as m_dt:
                m_dst.return_value = m_dst_res
                m_dt.return_value = m_dt_res
                res = self.cls._trans_dict(m)
        assert res == m_dt_res
        assert m_dt.mock_calls == [call(self.cls, m)]
        assert m_dst.mock_calls == []

    def test_scheduledtransaction(self):
        m = Mock(spec_set=ScheduledTransaction)
        m_dst_res = Mock()
        m_dt_res = Mock()
        with patch('%s._dict_for_sched_trans' % pb, autospec=True) as m_dst:
            with patch('%s._dict_for_trans' % pb, autospec=True) as m_dt:
                m_dst.return_value = m_dst_res
                m_dt.return_value = m_dt_res
                res = self.cls._trans_dict(m)
        assert res == m_dst_res
        assert m_dt.mock_calls == []
        assert m_dst.mock_calls == [call(self.cls, m, weekly_occurrence=0)]


class TestDictForTrans(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)

    def test_simple(self):
        m_account = Mock(name='foo')
        type(m_account).name = 'foo'
        m_budget = Mock(name='bar')
        type(m_budget).name = 'bar'
        m = Mock(
            spec=Transaction,
            id=123,
            date=date(year=2017, month=7, day=15),
            scheduled_trans_id=567,
            description='desc',
            actual_amount=Decimal('123.45'),
            budgeted_amount=Decimal('120.00'),
            account_id=2,
            account=m_account,
            planned_budget_id=3,
            planned_budget=m_budget,
            budget_transactions=[
                Mock(
                    spec_set=BudgetTransaction,
                    budget_id=3,
                    budget=m_budget,
                    amount=Decimal('123.45')
                )
            ]
        )
        type(m).reconcile = None
        assert self.cls._dict_for_trans(m) == {
            'type': 'Transaction',
            'id': 123,
            'date': date(year=2017, month=7, day=15),
            'sched_type': None,
            'sched_trans_id': 567,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': Decimal('120.00'),
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'planned_budget_id': 3,
            'planned_budget_name': 'bar',
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_budget_split(self):
        m_account = Mock(name='foo')
        type(m_account).name = 'foo'
        m_budget = Mock(name='bar')
        type(m_budget).name = 'bar'
        m_budget2 = Mock(name='foo')
        type(m_budget2).name = 'foo'
        m_budget3 = Mock(name='baz')
        type(m_budget3).name = 'baz'
        m = Mock(
            spec=Transaction,
            id=123,
            date=date(year=2017, month=7, day=15),
            scheduled_trans_id=567,
            description='desc',
            actual_amount=Decimal('123.45'),
            budgeted_amount=Decimal('120.00'),
            account_id=2,
            account=m_account,
            planned_budget_id=3,
            planned_budget=m_budget,
            budget_transactions=[
                Mock(
                    spec_set=BudgetTransaction,
                    budget_id=1,
                    budget=m_budget,
                    amount=Decimal('53.00')
                ),
                Mock(
                    spec_set=BudgetTransaction,
                    budget_id=2,
                    budget=m_budget2,
                    amount=Decimal('20.45')
                ),
                Mock(
                    spec_set=BudgetTransaction,
                    budget_id=3,
                    budget=m_budget3,
                    amount=Decimal('50.00')
                )
            ]
        )
        type(m).reconcile = None
        assert self.cls._dict_for_trans(m) == {
            'type': 'Transaction',
            'id': 123,
            'date': date(year=2017, month=7, day=15),
            'sched_type': None,
            'sched_trans_id': 567,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': Decimal('120.00'),
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'planned_budget_id': 3,
            'planned_budget_name': 'bar',
            'budgets': {
                1: {'name': 'bar', 'amount': Decimal('53.00')},
                2: {'name': 'foo', 'amount': Decimal('20.45')},
                3: {'name': 'baz', 'amount': Decimal('50.00')}
            }
        }

    def test_reconciled(self):
        m_account = Mock(name='foo')
        type(m_account).name = 'foo'
        m_budget = Mock(name='bar')
        type(m_budget).name = 'bar'
        m = Mock(
            spec=Transaction,
            id=123,
            date=date(year=2017, month=7, day=15),
            scheduled_trans_id=567,
            description='desc',
            actual_amount=Decimal('123.45'),
            budgeted_amount=Decimal('120.00'),
            account_id=2,
            account=m_account,
            planned_budget_id=3,
            planned_budget=m_budget,
            budget_transactions=[
                Mock(
                    spec_set=BudgetTransaction,
                    budget_id=3,
                    budget=m_budget,
                    amount=Decimal('123.45')
                )
            ]
        )
        type(m).reconcile = Mock(id=2)
        assert self.cls._dict_for_trans(m) == {
            'type': 'Transaction',
            'id': 123,
            'date': date(year=2017, month=7, day=15),
            'sched_type': None,
            'sched_trans_id': 567,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': Decimal('120.00'),
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': 2,
            'planned_budget_id': 3,
            'planned_budget_name': 'bar',
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_budgeted_amount_none(self):
        m_account = Mock(name='foo')
        type(m_account).name = 'foo'
        m_budget = Mock(name='bar')
        type(m_budget).name = 'bar'
        m = Mock(
            spec=Transaction,
            id=123,
            date=date(year=2017, month=7, day=15),
            scheduled_trans_id=567,
            description='desc',
            actual_amount=Decimal('123.45'),
            budgeted_amount=None,
            account_id=2,
            account=m_account,
            planned_budget_id=None,
            planned_budget=None,
            budget_transactions=[
                Mock(
                    spec_set=BudgetTransaction,
                    budget_id=3,
                    budget=m_budget,
                    amount=Decimal('123.45')
                )
            ]
        )
        type(m).reconcile = None
        assert self.cls._dict_for_trans(m) == {
            'type': 'Transaction',
            'id': 123,
            'date': date(year=2017, month=7, day=15),
            'sched_type': None,
            'sched_trans_id': 567,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'planned_budget_id': None,
            'planned_budget_name': None,
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }


class TestDictForSchedTrans(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)
        self.cls = BiweeklyPayPeriod(date(2017, 3, 7), self.mock_sess)
        m_account = Mock(name='foo')
        type(m_account).name = 'foo'
        m_budget = Mock(name='bar')
        type(m_budget).name = 'bar'
        self.m_st = Mock(
            spec_set=ScheduledTransaction,
            id=123,
            description='desc',
            amount=Decimal('123.45'),
            account_id=2,
            account=m_account,
            budget_id=3,
            budget=m_budget
        )

    def test_date(self):
        type(self.m_st).schedule_type = 'date'
        type(self.m_st).date = date(year=2017, month=7, day=15)
        assert self.cls._dict_for_sched_trans(self.m_st) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': date(year=2017, month=7, day=15),
            'sched_type': 'date',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_per_period(self):
        type(self.m_st).schedule_type = 'per period'
        assert self.cls._dict_for_sched_trans(self.m_st) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': None,
            'sched_type': 'per period',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_day_of_month_single_month(self):
        type(self.m_st).schedule_type = 'monthly'
        type(self.m_st).day_of_month = 9
        assert self.cls._dict_for_sched_trans(self.m_st) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': date(year=2017, month=3, day=9),
            'sched_type': 'monthly',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_day_of_month_cross_month_before(self):
        type(self.m_st).schedule_type = 'monthly'
        type(self.m_st).day_of_month = 28
        cls = BiweeklyPayPeriod(date(2017, 3, 26), self.mock_sess)
        assert cls._dict_for_sched_trans(self.m_st) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': date(year=2017, month=3, day=28),
            'sched_type': 'monthly',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_day_of_month_cross_month_after(self):
        type(self.m_st).schedule_type = 'monthly'
        type(self.m_st).day_of_month = 4
        cls = BiweeklyPayPeriod(date(2017, 3, 26), self.mock_sess)
        assert cls._dict_for_sched_trans(self.m_st) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': date(year=2017, month=4, day=4),
            'sched_type': 'monthly',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_weekly(self):
        type(self.m_st).schedule_type = 'weekly'
        type(self.m_st).day_of_week = 0  # Monday
        # Pay period 2017-03-07 to 2017-03-20
        # Mondays: 2017-03-13, 2017-03-20
        assert self.cls._dict_for_sched_trans(self.m_st, weekly_occurrence=0) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': date(year=2017, month=3, day=13),
            'sched_type': 'weekly',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_weekly_second_occurrence(self):
        type(self.m_st).schedule_type = 'weekly'
        type(self.m_st).day_of_week = 0  # Monday
        # Pay period 2017-03-07 to 2017-03-20
        # Mondays: 2017-03-13, 2017-03-20
        assert self.cls._dict_for_sched_trans(self.m_st, weekly_occurrence=1) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': date(year=2017, month=3, day=20),
            'sched_type': 'weekly',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_annual(self):
        type(self.m_st).schedule_type = 'annual'
        type(self.m_st).annual_month = 3
        type(self.m_st).annual_day = 15
        # Pay period 2017-03-07 to 2017-03-20
        assert self.cls._dict_for_sched_trans(self.m_st) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': date(year=2017, month=3, day=15),
            'sched_type': 'annual',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }

    def test_annual_not_in_period(self):
        type(self.m_st).schedule_type = 'annual'
        type(self.m_st).annual_month = 4
        type(self.m_st).annual_day = 15
        # Pay period 2017-03-07 to 2017-03-20, April 15 is not in period
        assert self.cls._dict_for_sched_trans(self.m_st) == {
            'type': 'ScheduledTransaction',
            'id': 123,
            'date': None,
            'sched_type': 'annual',
            'sched_trans_id': None,
            'description': 'desc',
            'amount': Decimal('123.45'),
            'budgeted_amount': None,
            'account_id': 2,
            'account_name': 'foo',
            'reconcile_id': None,
            'budgets': {
                3: {'name': 'bar', 'amount': Decimal('123.45')}
            }
        }


class TestSTWeekly(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)

    def test_scheduled_transactions_weekly(self):
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        res = cls._scheduled_transactions_weekly()
        qfr = self.mock_sess.query.return_value.filter.return_value
        assert res == qfr.order_by.return_value
        assert len(self.mock_sess.mock_calls) == 3
        assert self.mock_sess.mock_calls[0] == call.query(ScheduledTransaction)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected = [
            ScheduledTransaction.schedule_type.__eq__('weekly'),
            ScheduledTransaction.is_active.__eq__(True)
        ]
        for idx, exp in enumerate(expected):
            assert str(kall[1][idx]) == str(exp)
        kall = self.mock_sess.mock_calls[2]
        assert kall[0] == 'query().filter().order_by'
        assert str(kall[1][0]) == str(asc(ScheduledTransaction.day_of_week))
        assert str(kall[1][1]) == str(asc(ScheduledTransaction.amount))


class TestSTAnnual(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)

    def test_scheduled_transactions_annual(self):
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        res = cls._scheduled_transactions_annual()
        assert res == self.mock_sess.query.return_value.filter.return_value
        assert len(self.mock_sess.mock_calls) == 2
        assert self.mock_sess.mock_calls[0] == call.query(ScheduledTransaction)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected = [
            ScheduledTransaction.schedule_type.__eq__('annual'),
            ScheduledTransaction.is_active.__eq__(True)
        ]
        for idx, exp in enumerate(expected):
            assert str(kall[1][idx]) == str(exp)


class TestFilterAnnualForPeriod(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)

    def test_annual_in_period(self):
        # Pay period 2017-03-17 to 2017-03-30
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        m_trans = Mock(spec_set=ScheduledTransaction, annual_month=3,
                       annual_day=20)
        result = cls._filter_annual_for_period([m_trans])
        assert result == [m_trans]

    def test_annual_not_in_period(self):
        # Pay period 2017-03-17 to 2017-03-30
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        m_trans = Mock(spec_set=ScheduledTransaction, annual_month=4,
                       annual_day=15)
        result = cls._filter_annual_for_period([m_trans])
        assert result == []

    def test_annual_at_period_start(self):
        # Pay period 2017-03-17 to 2017-03-30
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        m_trans = Mock(spec_set=ScheduledTransaction, annual_month=3,
                       annual_day=17)
        result = cls._filter_annual_for_period([m_trans])
        assert result == [m_trans]

    def test_annual_at_period_end(self):
        # Pay period 2017-03-17 to 2017-03-30
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        m_trans = Mock(spec_set=ScheduledTransaction, annual_month=3,
                       annual_day=30)
        result = cls._filter_annual_for_period([m_trans])
        assert result == [m_trans]

    def test_annual_multiple_transactions(self):
        # Pay period 2017-03-17 to 2017-03-30
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        m_trans1 = Mock(spec_set=ScheduledTransaction, annual_month=3,
                        annual_day=20)
        m_trans2 = Mock(spec_set=ScheduledTransaction, annual_month=4,
                        annual_day=15)  # Not in period
        m_trans3 = Mock(spec_set=ScheduledTransaction, annual_month=3,
                        annual_day=25)
        result = cls._filter_annual_for_period([m_trans1, m_trans2, m_trans3])
        assert result == [m_trans1, m_trans3]

    def test_feb_29_in_leap_year(self):
        # Pay period in a leap year that includes Feb 29
        # 2020-02-21 to 2020-03-05
        cls = BiweeklyPayPeriod(date(2020, 2, 21), self.mock_sess)
        m_trans = Mock(spec_set=ScheduledTransaction, annual_month=2,
                       annual_day=29)
        result = cls._filter_annual_for_period([m_trans])
        assert result == [m_trans]

    def test_feb_29_in_non_leap_year(self):
        # Pay period in a non-leap year that would include Feb 29
        # 2017-02-17 to 2017-03-02
        cls = BiweeklyPayPeriod(date(2017, 2, 17), self.mock_sess)
        m_trans = Mock(spec_set=ScheduledTransaction, annual_month=2,
                       annual_day=29)
        result = cls._filter_annual_for_period([m_trans])
        # Should be empty because Feb 29 doesn't exist in 2017
        assert result == []


class TestDatesForWeekdayInPeriod(object):

    def setup_method(self):
        self.mock_sess = Mock(spec_set=Session)

    def test_monday(self):
        # Pay period 2017-03-17 (Friday) to 2017-03-30 (Thursday)
        # Mondays: 2017-03-20, 2017-03-27
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        result = cls._dates_for_weekday_in_period(0)  # Monday
        assert result == [date(2017, 3, 20), date(2017, 3, 27)]

    def test_friday(self):
        # Pay period 2017-03-17 (Friday) to 2017-03-30 (Thursday)
        # Fridays: 2017-03-17, 2017-03-24
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        result = cls._dates_for_weekday_in_period(4)  # Friday
        assert result == [date(2017, 3, 17), date(2017, 3, 24)]

    def test_sunday(self):
        # Pay period 2017-03-17 (Friday) to 2017-03-30 (Thursday)
        # Sundays: 2017-03-19, 2017-03-26
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        result = cls._dates_for_weekday_in_period(6)  # Sunday
        assert result == [date(2017, 3, 19), date(2017, 3, 26)]

    def test_always_two_occurrences(self):
        # Every weekday should have exactly 2 occurrences in a 14-day period
        cls = BiweeklyPayPeriod(date(2017, 3, 17), self.mock_sess)
        for day_of_week in range(7):
            result = cls._dates_for_weekday_in_period(day_of_week)
            assert len(result) == 2, \
                f"Expected 2 occurrences for weekday {day_of_week}"



from __future__ import absolute_import
from __future__ import unicode_literals
from biweeklybudget.vendored.ofxclient.account import (
    Account, BrokerageAccount, CreditCardAccount, BankAccount
)
from biweeklybudget.vendored.ofxclient.client import Client
from biweeklybudget.vendored.ofxclient.institution import Institution
from biweeklybudget.vendored.ofxclient.version import __version__



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from datetime import timedelta, datetime, date
from functools import total_ordering
from sqlalchemy import or_, asc
from dateutil import relativedelta
from collections import defaultdict
from decimal import Decimal

from biweeklybudget import settings
from biweeklybudget.models import Transaction, ScheduledTransaction, Budget
from biweeklybudget.utils import dtnow


@total_ordering
class BiweeklyPayPeriod(object):
    """
    This object contains all logic related to working with pay periods,
    specifically finding a pay period for a given data, and figuring out the
    start and end dates of pay periods. Sure, the app is called "biweeklybudget"
    but there's no reason to hard-code logic all over the place that's this
    simple.
    """

    def __init__(self, start_date, db_session):
        """
        Create a new BiweeklyPayPeriod instance.

        :param start_date: starting date of the pay period
        :type start_date: :py:class:`datetime.date` or
          :py:class:`datetime.datetime`
        :param db_session: active database session to use for queries
        :type db_session: sqlalchemy.orm.session.Session
        """
        if isinstance(start_date, datetime):
            start_date = start_date.date()
        self._db = db_session
        self._start_date = start_date
        self._end_date = start_date + self.period_length
        self._data_cache = {}
        self._income_budget_id_list = None

    @property
    def period_interval(self):
        """
        Return the interval between BiweeklyPayPeriods as a timedelta.

        :return: interval between BiweeklyPayPeriods
        :rtype: datetime.timedelta
        """
        return timedelta(days=14)

    @property
    def period_length(self):
        """
        Return the length of a BiweeklyPayPeriod; this is calculated as
        :py:attr:`~.period_interval` minus one second.

        :return: length of one BiweeklyPayPeriod
        :rtype: datetime.timedelta
        """
        return self.period_interval - timedelta(days=1)

    @property
    def start_date(self):
        """
        Return the starting date for this pay period. The period is generally
        considered to start at midnight (00:00) of this date.

        :return: start date for pay period
        :rtype: datetime.date
        """
        return self._start_date

    @property
    def end_date(self):
        """
        Return the date of the last day in this pay period. The pay period is
        generally considered to end at the last instant (i.e. 23:59:59) of this
        date.

        :return: last date in the pay period
        :rtype: datetime.date
        """
        return self._end_date

    @property
    def is_in_past(self):
        return self.end_date < dtnow().date()

    @property
    def next(self):
        """
        Return the BiweeklyPayPeriod following this one.

        :return: next BiweeklyPayPeriod after this one
        :rtype: BiweeklyPayPeriod
        """
        return BiweeklyPayPeriod(
            (self.start_date + self.period_interval),
            self._db
        )

    @property
    def previous(self):
        """
        Return the BiweeklyPayPeriod preceding this one.

        :return: previous BiweeklyPayPeriod before this one
        :rtype: BiweeklyPayPeriod
        """
        return BiweeklyPayPeriod(
            (self.start_date - self.period_interval),
            self._db
        )

    def __repr__(self):
        return '<BiweeklyPayPeriod(%s)>' % self._start_date.strftime('%Y-%m-%d')

    def __eq__(self, other):
        if not isinstance(other, BiweeklyPayPeriod):
            return NotImplemented
        return self.start_date == other.start_date

    def __lt__(self, other):
        if not isinstance(other, BiweeklyPayPeriod):
            return NotImplemented
        return self.start_date < other.start_date

    @staticmethod
    def period_for_date(dt, db_session):
        """
        Given a datetime, return the BiweeklyPayPeriod instance describing the
        pay period containing this date.

        .. todo:: This is a very naive, poorly-performing implementation.

        :param dt: datetime or date to find the pay period for
        :type dt: :py:class:`~datetime.datetime` or :py:class:`~datetime.date`
        :param db_session: active database session to use for queries
        :type db_session: sqlalchemy.orm.session.Session
        :return: BiweeklyPayPeriod containing the specified date
        :rtype: :py:class:`~.BiweeklyPayPeriod`
        """
        p = BiweeklyPayPeriod(settings.PAY_PERIOD_START_DATE, db_session)
        if isinstance(dt, datetime):
            dt = dt.date()
        if dt < p.start_date:
            while True:
                if p.end_date >= dt >= p.start_date:
                    return p
                p = p.previous
        if dt > p.end_date:
            while True:
                if p.end_date >= dt >= p.start_date:
                    return p
                p = p.next
        return p

    def filter_query(self, query, date_prop):
        """
        Filter ``query`` for ``date_prop`` in this pay period. Returns a copy
        of the query.

        e.g. to filter an existing query of :py:class:`~.OFXTransaction` for
        the BiweeklyPayPeriod starting on 2017-01-14:

        .. code-block:: python

            q = # some query here
            p = BiweeklyPayPeriod(date(2017, 1, 14))
            q = p.filter_query(q, OFXTransaction.date_posted)

        :param query: The query to filter
        :type query: ``sqlalchemy.orm.query.Query``
        :param date_prop: the Model's date property, to filter on.
        :return: the filtered query
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        return query.filter(
            date_prop >= self.start_date, date_prop <= self.end_date
        )

    def _transactions(self):
        """
        Return a Query for all :py:class:`~.Transaction` for this pay period.

        :return: Query matching all Transactions for this pay period
        :rtype: sqlalchemy.orm.query.Query
        """
        return self.filter_query(
            self._db.query(Transaction),
            Transaction.date
        )

    def _scheduled_transactions_date(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        date (schedule_type == "date") for this pay period.

        :return: Query matching all ScheduledTransactions defined by date, for
          this pay period.
        :rtype: sqlalchemy.orm.query.Query
        """
        return self.filter_query(
            self._db.query(ScheduledTransaction).filter(
                ScheduledTransaction.is_active.__eq__(True)),
            ScheduledTransaction.date
        )

    def _scheduled_transactions_per_period(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        number per period (schedule_type == "per period") for this pay period.

        :return: Query matching all ScheduledTransactions defined by number
          per period, for this pay period.
        :rtype: sqlalchemy.orm.query.Query
        """
        return self._db.query(ScheduledTransaction).filter(
            ScheduledTransaction.schedule_type.__eq__('per period'),
            ScheduledTransaction.is_active.__eq__(True)
        ).order_by(
            asc(ScheduledTransaction.num_per_period),
            asc(ScheduledTransaction.amount)
        )

    def _scheduled_transactions_monthly(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        day of month (schedule_type == "monthly") for this pay period.

        :return: Query matching all ScheduledTransactions defined by day of
          month (monthly) for this period.
        :rtype: sqlalchemy.orm.query.Query
        """
        if self.start_date.day < self.end_date.day:
            # start and end dates are contiguous, in the same month
            return self._db.query(ScheduledTransaction).filter(
                ScheduledTransaction.schedule_type.__eq__('monthly'),
                ScheduledTransaction.is_active.__eq__(True),
                ScheduledTransaction.day_of_month <= self.end_date.day,
                ScheduledTransaction.day_of_month >= self.start_date.day
            )
        # else we span two months
        return self._db.query(ScheduledTransaction).filter(
            ScheduledTransaction.schedule_type.__eq__('monthly'),
            ScheduledTransaction.is_active.__eq__(True),
            or_(
                ScheduledTransaction.day_of_month <= self.end_date.day,
                ScheduledTransaction.day_of_month >= self.start_date.day
            )
        )

    def _scheduled_transactions_weekly(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        day of week (schedule_type == "weekly"). Weekly transactions always
        occur in every pay period since each weekday appears exactly twice
        in a 14-day biweekly period.

        :return: Query matching all active weekly ScheduledTransactions.
        :rtype: sqlalchemy.orm.query.Query
        """
        return self._db.query(ScheduledTransaction).filter(
            ScheduledTransaction.schedule_type.__eq__('weekly'),
            ScheduledTransaction.is_active.__eq__(True)
        ).order_by(
            asc(ScheduledTransaction.day_of_week),
            asc(ScheduledTransaction.amount)
        )

    def _scheduled_transactions_annual(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        annual month and day (schedule_type == "annual") that fall within
        this pay period.

        :return: Query matching all active annual ScheduledTransactions for
          this pay period.
        :rtype: sqlalchemy.orm.query.Query
        """
        # We need to find annual transactions where the month/day falls in
        # [start_date, end_date]. This is complex because we're comparing
        # month+day without year.
        # We'll fetch all annual transactions and filter in Python.
        return self._db.query(ScheduledTransaction).filter(
            ScheduledTransaction.schedule_type.__eq__('annual'),
            ScheduledTransaction.is_active.__eq__(True)
        )

    def _filter_annual_for_period(self, annual_transactions):
        """
        Filter annual transactions to only those that fall within this pay
        period. This handles the complexity of month/day comparison across
        year boundaries.

        :param annual_transactions: list of annual ScheduledTransactions
        :return: list of annual ScheduledTransactions in this pay period
        :rtype: list
        """
        result = []
        for t in annual_transactions:
            try:
                # Try to construct the date for this year
                annual_date = date(
                    year=self.start_date.year,
                    month=t.annual_month,
                    day=t.annual_day
                )
            except ValueError:
                # Invalid date (e.g., Feb 29 in non-leap year)
                # Try next year in case period spans year boundary
                try:
                    annual_date = date(
                        year=self.start_date.year + 1,
                        month=t.annual_month,
                        day=t.annual_day
                    )
                except ValueError:
                    # Still invalid, skip this transaction
                    continue

            # Check if this date falls in our pay period
            if self.start_date <= annual_date <= self.end_date:
                result.append(t)
                continue

            # If the pay period spans a year boundary, check the next year too
            if self.end_date.year > self.start_date.year:
                try:
                    annual_date_next = date(
                        year=self.end_date.year,
                        month=t.annual_month,
                        day=t.annual_day
                    )
                    if self.start_date <= annual_date_next <= self.end_date:
                        result.append(t)
                except ValueError:
                    # Invalid date, skip
                    pass
        return result

    @property
    def transactions_list(self):
        """
        Return an ordered list of dicts, each representing a transaction for
        this pay period. Dicts have keys and values as described in
        :py:meth:`~._trans_dict`.

        :return: ordered list of transaction dicts
        :rtype: list
        """
        return self._data['all_trans_list']

    @property
    def _income_budget_ids(self):
        """
        Return a list of all :py:class:`~.Budget` IDs for Income budgets.

        :return: list of income budget IDs
        :rtype: list
        """
        if self._income_budget_id_list is None:
            self._income_budget_id_list = [
                b.id for b in self._db.query(
                    Budget).filter(Budget.is_income.__eq__(True)).all()
            ]
        return self._income_budget_id_list

    @property
    def _data(self):
        """
        Return the object-local data cache dict. Build it if not already
        present.

        :return: object-local data cache
        :rtype: dict
        """
        if len(self._data_cache) > 0:
            return self._data_cache
        self._data_cache = {
            'transactions': self._transactions().all(),
            'st_date': self._scheduled_transactions_date().all(),
            'st_per_period': self._scheduled_transactions_per_period().all(),
            'st_monthly': self._scheduled_transactions_monthly().all(),
            'st_weekly': self._scheduled_transactions_weekly().all(),
            'st_annual': self._filter_annual_for_period(
                self._scheduled_transactions_annual().all()
            )
        }
        self._data_cache['all_trans_list'] = self._make_combined_transactions()
        self._data_cache['budget_sums'] = self._make_budget_sums()
        self._data_cache['overall_sums'] = self._make_overall_sums()
        return self._data_cache

    def clear_cache(self):
        """
        Clear the cached transaction, budget and sum data stored in
        `self._data_cache` and returned by :py:attr:`~._data`.
        """
        self._data_cache = {}

    def _make_combined_transactions(self):
        """
        Combine all Transactions and ScheduledTransactions from
        ``self._data_cache`` into one ordered list of similar dicts, adding
        dates to the monthly ScheduledTransactions as appropriate and excluding
        ScheduledTransactions that have been converted to real Transactions.
        Store the finished list back into ``self._data_cache``.
        """
        unordered = []
        # ScheduledTransaction ID to count of real trans for each
        st_ids = defaultdict(int)
        for t in self._data_cache['transactions']:
            unordered.append(self._trans_dict(t))
            if t.scheduled_trans_id is not None:
                st_ids[t.scheduled_trans_id] += 1
        for t in self._data_cache['st_date']:
            if t.id not in st_ids:
                unordered.append(self._trans_dict(t))
        for t in self._data_cache['st_monthly']:
            if t.id not in st_ids:
                unordered.append(self._trans_dict(t))
        for t in self._data_cache['st_annual']:
            if t.id not in st_ids:
                unordered.append(self._trans_dict(t))
        ordered = []
        for t in self._data_cache['st_per_period']:
            d = self._trans_dict(t)
            for _ in range(0, (t.num_per_period - st_ids[t.id])):
                ordered.append(d)
        # Weekly transactions appear twice per pay period (14 days = 2 weeks)
        for t in self._data_cache['st_weekly']:
            remaining = 2 - st_ids[t.id]
            for occ_idx in range(remaining):
                d = self._trans_dict(t, weekly_occurrence=occ_idx)
                ordered.append(d)
        for t in sorted(unordered, key=lambda k: k['date']):
            ordered.append(t)

        def sortkey(k):
            d = k.get('date', None)
            if d is None:
                d = date.min
            return d, k['amount']

        return sorted(ordered, key=sortkey)

    @property
    def budget_sums(self):
        """
        Return a dict of budget sums; the return value of
        :py:meth:`~._make_budget_sums`.

        :return: dict of dicts, transaction sums and amounts per budget
        :rtype: dict
        """
        return self._data['budget_sums']

    def _make_budget_sums(self):
        """
        Find the sums of all transactions per periodic budget ID ; return a dict
        where keys are budget IDs and values are per-budget dicts containing:

        - ``budget_amount`` *(Decimal.decimal)* - the periodic budget
          :py:attr:`~.Budget.starting_balance`.
        - ``allocated`` *(Decimal.decimal)* - sum of all
          :py:class:`~.ScheduledTransaction` and :py:class:`~.Transaction`
          amounts against the budget this period. For actual transactions, we
          use the :py:attr:`~.Transaction.budgeted_amount` if present (not
          None).
        - ``spent`` *(Decimal.decimal)* - the sum of all actual
          :py:class:`~.Transaction` amounts against the budget this period.
        - ``trans_total`` *(Decimal.decimal)* - the sum of spent amounts for
          Transactions that have them, or allocated amounts for
          ScheduledTransactions.
        - ``remaining`` *(Decimal.decimal)* - the remaining amount in the
          budget. This is ``budget_amount`` minus the greater of ``allocated``
          or ``trans_total``. For income budgets, this is always positive.

        :return: dict of dicts, transaction sums and amounts per budget
        :rtype: dict
        """
        res = {}
        for b in self._db.query(Budget).filter(
            Budget.is_active.__eq__(True),
            Budget.is_periodic.__eq__(True)
        ).all():
            res[b.id] = {
                'budget_amount': b.starting_balance,
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': b.is_income
            }
        for t in self.transactions_list:
            # if a ScheduledTransaction, update some values and then continue
            if t['type'] == 'ScheduledTransaction':
                for budg_id, budg_data in t['budgets'].items():
                    if budg_id not in res:
                        # Issue #161 - inactive budget, but transaction for it
                        # in this period. Add it if it's a periodic budget.
                        b = self._db.query(Budget).get(budg_id)
                        if not b.is_periodic:
                            continue
                        res[b.id] = {
                            'budget_amount': b.starting_balance,
                            'allocated': Decimal('0.0'),
                            'spent': Decimal('0.0'),
                            'trans_total': Decimal('0.0'),
                            'is_income': b.is_income
                        }
                    res[budg_id]['allocated'] += budg_data['amount']
                    res[budg_id]['trans_total'] += budg_data['amount']
                continue
            # NOT a ScheduledTransaction; must be an actual Transaction
            for budg_id, budg_data in t['budgets'].items():
                if budg_id not in res:
                    # Issue #161 - inactive budget, but transaction for it
                    # in this period. Add it if it's a periodic budget.
                    b = self._db.query(Budget).get(budg_id)
                    if not b.is_periodic:
                        continue
                    res[b.id] = {
                        'budget_amount': b.starting_balance,
                        'allocated': Decimal('0.0'),
                        'spent': Decimal('0.0'),
                        'trans_total': Decimal('0.0'),
                        'is_income': b.is_income
                    }
                # update the budget's transactions total and spent amount
                res[budg_id]['trans_total'] += budg_data['amount']
                res[budg_id]['spent'] += budg_data['amount']
                if t['budgeted_amount'] is None:
                    res[budg_id]['allocated'] += budg_data['amount']
            if t['budgeted_amount'] is not None:
                # has a budgeted amount. It _shouldn't_ be possible to have a
                # budgeted_amount without a planned_budget_id, but if that
                # happens, allocate to the first budget.
                if t.get('planned_budget_id', None) is None:
                    bid = list(t['budgets'].keys())[0]
                else:
                    bid = t['planned_budget_id']
                if bid not in res:
                    # Issue #161 - inactive budget, but transaction for it
                    # in this period. Add it if it's a periodic budget.
                    b = self._db.query(Budget).get(bid)
                    if not b.is_periodic:
                        continue
                    res[bid] = {
                        'budget_amount': b.starting_balance,
                        'allocated': Decimal('0.0'),
                        'spent': Decimal('0.0'),
                        'trans_total': Decimal('0.0'),
                        'is_income': b.is_income
                    }
                res[bid]['allocated'] += t['budgeted_amount']
        for b in res.keys():
            if res[b]['trans_total'] > res[b]['allocated']:
                res[b]['remaining'] = res[
                    b]['budget_amount'] - res[b]['trans_total']
            else:
                res[b]['remaining'] = res[
                    b]['budget_amount'] - res[b]['allocated']
            if res[b]['is_income']:
                res[b]['remaining'] = abs(res[b]['remaining'])
        return res

    @property
    def overall_sums(self):
        """
        Return a dict of overall sums; the return value of
        :py:meth:`~._make_overall_sums`.

        :return: dict describing sums for the pay period
        :rtype: dict
        """
        return self._data['overall_sums']

    def _make_overall_sums(self):
        """
        Return a dict describing the overall sums for this pay period, namely:

        - ``allocated`` *(Decimal.decimal)* total amount allocated via
          :py:class:`~.ScheduledTransaction`,
          :py:class:`~.Transaction` (counting the
          :py:attr:`~.Transaction.budgeted_amount` for Transactions that have
          one), or :py:class:`~.Budget` (not counting income budgets).
        - ``spent`` *(Decimal.decimal)* total amount actually spent via
          :py:class:`~.Transaction`.
        - ``income`` *(Decimal.decimal)* total amount of income allocated this
          pay period. Calculated value (from :py:meth:`~._make_budget_sums` /
          ``self._data_cache['budget_sums']``) should be negative, but is
          returned as its positive inverse (absolute value).
        - ``remaining`` *(Decimal.decimal)* income minus the greater of
          ``allocated`` or ``spent`` for current or future pay periods, or minus
          ``spent`` for pay periods ending in the past (:py:attr:`~.is_in_past`)

        :return: dict describing sums for the pay period
        :rtype: dict
        """
        budgets_total = Decimal('0.0')
        res = {
            'allocated': Decimal('0.0'),
            'spent': Decimal('0.0'),
            'income': Decimal('0.0'),
            'remaining': Decimal('0.0')
        }
        for _, b in self._data_cache['budget_sums'].items():
            if b['is_income']:
                if abs(b['trans_total']) > abs(b['budget_amount']):
                    res['income'] += abs(b['trans_total'])
                else:
                    res['income'] += abs(b['budget_amount'])
                continue
            res['allocated'] += max(b['allocated'], b['budget_amount'])
            res['spent'] += b['spent']
            budgets_total += max(b['trans_total'], b['budget_amount'])
        if self.is_in_past:
            res['remaining'] = res['income'] - res['spent']
        else:
            res['remaining'] = res['income'] - budgets_total
        return res

    def _trans_dict(self, t, weekly_occurrence=0):
        """
        Given a Transaction or ScheduledTransaction, return a dict of a
        common format describing the object.

        The resulting dict will have the following layout:

        * ``type`` (**str**) "Transaction" or "ScheduledTransaction"
        * ``id`` (**int**) the id of the object
        * ``date`` (**date**) the date of the transaction, or None for
          per-period ScheduledTransactions
        * ``sched_type`` (**str**) for ScheduledTransactions, the schedule
          type ("monthly", "date", "per period", "weekly", or "annual")
        * ``sched_trans_id`` (**int**) for Transactions, the
          ScheduledTransaction ``id`` that it was created from, or None.
        * ``description`` (**str**) the transaction description
        * ``amount`` (**Decimal.decimal**) the transaction amount
        * ``budgeted_amount`` (**Decimal.decimal**) the budgeted amount.
          This may be None.
        * ``account_id`` (**int**) the id of the Account the transaction is
          against.
        * ``account_name`` (**str**) the name of the Account the transaction is
          against.
        * ``budgets`` (**dict**) dict of information on the Budgets this
          Transaction is against. Keys are budget IDs (**int**), values are
          dicts with keys "amount" (**Decimal**) and "name" (**string**).
        * ``reconcile_id`` (**int**) the ID of the TxnReconcile, or None
        * ``planned_budget_id`` (**int**) the id of the Budget the transaction
          was planned against, if any. May be None.
        * ``planned_budget_name`` (**str**) the name of the Budget the
          transaction was planned against, if any. May be None.

        :param t: the object to return a dict for
        :type t: :py:class:`~.Transaction` or :py:class:`~.ScheduledTransaction`
        :param weekly_occurrence: For weekly transactions, which occurrence
          (0 or 1) within the pay period this dict is for.
        :type weekly_occurrence: int
        :return: dict describing ``t``
        :rtype: dict
        """
        if isinstance(t, Transaction):
            return self._dict_for_trans(t)
        return self._dict_for_sched_trans(t, weekly_occurrence=weekly_occurrence)

    def _dict_for_trans(self, t):
        """
        Return a dict describing the Transaction t. Called from
        :py:meth:`~._trans_dict`.

        The resulting dict will have the following layout:

        * ``type`` (**str**) "Transaction" or "ScheduledTransaction"
        * ``id`` (**int**) the id of the object
        * ``date`` (**date**) the date of the transaction, or None for
          per-period ScheduledTransactions
        * ``sched_type`` (**str**) for ScheduledTransactions, the schedule
          type ("monthly", "date", or "per period")
        * ``sched_trans_id`` (**int**) for Transactions, the
          ScheduledTransaction ``id`` that it was created from, or None.
        * ``description`` (**str**) the transaction description
        * ``amount`` (**Decimal.decimal**) the transaction amount
        * ``budgeted_amount`` (**Decimal.decimal**) the budgeted amount.
          This may be None.
        * ``account_id`` (**int**) the id of the Account the transaction is
          against.
        * ``account_name`` (**str**) the name of the Account the transaction is
          against.
        * ``reconcile_id`` (**int**) the ID of the TxnReconcile, or None
        * ``planned_budget_id`` (**int**) the id of the Budget the transaction
          was planned against, if any. May be None.
        * ``planned_budget_name`` (**str**) the name of the Budget the
          transaction was planned against, if any. May be None.
        * ``budgets`` (**dict**) dict of information on the Budgets this
          Transaction is against. Keys are budget IDs (**int**), values are
          dicts with keys "amount" (**Decimal**) and "name" (**string**).

        :param t: transaction to describe
        :type t: Transaction
        :return: common-format dict describing ``t``
        :rtype: dict
        """
        res = {
            'type': 'Transaction',
            'id': t.id,
            'date': t.date,
            'sched_type': None,
            'sched_trans_id': t.scheduled_trans_id,
            'description': t.description,
            'amount': t.actual_amount,
            'account_id': t.account_id,
            'account_name': t.account.name,
            'planned_budget_id': t.planned_budget_id,
            'planned_budget_name': None,
            'budgets': {
                bt.budget_id: {
                    'amount': bt.amount,
                    'name': bt.budget.name
                } for bt in t.budget_transactions
            }
        }
        if t.reconcile is None:
            res['reconcile_id'] = None
        else:
            res['reconcile_id'] = t.reconcile.id
        if t.budgeted_amount is None:
            res['budgeted_amount'] = None
        else:
            res['budgeted_amount'] = t.budgeted_amount
        if t.planned_budget is not None:
            res['planned_budget_name'] = t.planned_budget.name
        return res

    def _dict_for_sched_trans(self, t, weekly_occurrence=0):
        """
        Return a dict describing the ScheduledTransaction t. Called from
        :py:meth:`~._trans_dict`.

        The resulting dict will have the following layout:

        * ``type`` (**str**) "Transaction" or "ScheduledTransaction"
        * ``id`` (**int**) the id of the object
        * ``date`` (**date**) the date of the transaction, or None for
          per-period ScheduledTransactions
        * ``sched_type`` (**str**) for ScheduledTransactions, the schedule
          type ("monthly", "date", "per period", "weekly", or "annual")
        * ``sched_trans_id`` **None**
        * ``description`` (**str**) the transaction description
        * ``amount`` (**Decimal.decimal**) the transaction amount
        * ``budgeted_amount`` **None**
        * ``account_id`` (**int**) the id of the Account the transaction is
          against.
        * ``account_name`` (**str**) the name of the Account the transaction is
          against.
        * ``reconcile_id`` (**int**) the ID of the TxnReconcile, or None
        * ``budgets`` (**dict**) dict of information on the Budgets this
          Transaction is against. Keys are budget IDs (**int**), values are
          dicts with keys "amount" (**Decimal**) and "name" (**string**).

        :param t: ScheduledTransaction to describe
        :type t: ScheduledTransaction
        :param weekly_occurrence: For weekly transactions, which occurrence
          (0 or 1) within the pay period this dict is for.
        :type weekly_occurrence: int
        :return: common-format dict describing ``t``
        :rtype: dict
        """
        res = {
            'type': 'ScheduledTransaction',
            'id': t.id,
            'sched_type': t.schedule_type,
            'sched_trans_id': None,
            'description': t.description,
            'amount': t.amount,
            'budgeted_amount': None,
            'account_id': t.account_id,
            'account_name': t.account.name,
            'reconcile_id': None,
            'budgets': {
                t.budget_id: {
                    'name': t.budget.name,
                    'amount': t.amount
                }
            }
        }
        if t.schedule_type == 'date':
            res['date'] = t.date
            return res
        if t.schedule_type == 'per period':
            res['date'] = None
            return res
        if t.schedule_type == 'weekly':
            # Find the dates in this pay period that fall on the specified
            # weekday. There are exactly 2 occurrences in a 14-day period.
            dates_for_weekday = self._dates_for_weekday_in_period(t.day_of_week)
            if weekly_occurrence < len(dates_for_weekday):
                res['date'] = dates_for_weekday[weekly_occurrence]
            else:
                res['date'] = None
            return res
        if t.schedule_type == 'annual':
            # Calculate the annual date for this period
            try:
                annual_date = date(
                    year=self.start_date.year,
                    month=t.annual_month,
                    day=t.annual_day
                )
                if self.start_date <= annual_date <= self.end_date:
                    res['date'] = annual_date
                    return res
            except ValueError:
                pass
            # Try next year if period spans year boundary
            if self.end_date.year > self.start_date.year:
                try:
                    annual_date = date(
                        year=self.end_date.year,
                        month=t.annual_month,
                        day=t.annual_day
                    )
                    if self.start_date <= annual_date <= self.end_date:
                        res['date'] = annual_date
                        return res
                except ValueError:
                    pass
            res['date'] = None
            return res
        # else it's a monthly transaction, and we need to figure out the date
        # for it that falls in this PayPeriod.
        if self.start_date.day <= t.day_of_month <= self.end_date.day:
            res['date'] = date(
                year=self.start_date.year,
                month=self.start_date.month,
                day=t.day_of_month
            )
            return res
        # If we got here, we're in a pay period that spans two months; i.e.
        # start_date.day > end_date.day.
        if t.day_of_month >= self.start_date.day:
            # in the same month as start_date.day
            res['date'] = date(
                year=self.start_date.year,
                month=self.start_date.month,
                day=t.day_of_month
            )
            return res
        # else t.day_of_month < self.start_date.day, which means it's actually
        # ``t.day_of_month`` in the next month...
        res['date'] = date(
            year=self.start_date.year,
            month=self.start_date.month,
            day=t.day_of_month
        ) + relativedelta.relativedelta(months=1)
        return res

    def _dates_for_weekday_in_period(self, day_of_week):
        """
        Return a list of dates in this pay period that fall on the specified
        weekday.

        :param day_of_week: The day of week (0=Monday, 6=Sunday)
        :type day_of_week: int
        :return: List of dates (exactly 2 for a 14-day biweekly period)
        :rtype: list of date
        """
        result = []
        current = self.start_date
        while current <= self.end_date:
            if current.weekday() == day_of_week:
                result.append(current)
            current += timedelta(days=1)
        return result



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""
import sys
from decimal import Decimal
from datetime import date
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import null

from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.budget_transaction import BudgetTransaction
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.tests.unit_helpers import binexp_to_dict

import pytest

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import Mock, patch, call  # noqa
else:
    from unittest.mock import Mock, patch, call  # noqa

pbm = 'biweeklybudget.models.transaction'


class TestTransactionModel(object):

    @patch('%s.RECONCILE_BEGIN_DATE' % pbm, date(2017, 3, 17))
    def test_unreconciled(self):
        Transaction()
        m_db = Mock()
        m_q = Mock(spec_set=Query)
        m_filt = Mock(spec_set=Query)
        m_db.query.return_value = m_q
        m_q.filter.return_value = m_filt
        res = Transaction.unreconciled(m_db)
        assert res == m_filt
        assert len(m_db.mock_calls) == 2
        assert m_db.mock_calls[0] == call.query(Transaction)
        kall = m_db.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected1 = Transaction.reconcile.__eq__(null())
        expected2 = Transaction.date.__ge__(date(2017, 3, 17))
        expected3 = Transaction.account.has(reconcile_trans=True)
        assert len(kall[1]) == 3
        assert str(expected1) == str(kall[1][0])
        assert binexp_to_dict(expected2) == binexp_to_dict(kall[1][1])
        assert str(expected3) == str(kall[1][2])


class TestSetBudgetAmounts(object):

    def test_none(self):
        mock_sess = Mock()
        with patch('%s.inspect' % pbm) as mock_inspect:
            type(mock_inspect.return_value).session = mock_sess
            t = Transaction()
            with pytest.raises(AssertionError):
                t.set_budget_amounts({})
        assert mock_sess.mock_calls == []

    def test_none_but_existing(self):
        mock_sess = Mock()
        with patch('%s.inspect' % pbm) as mock_inspect:
            type(mock_inspect.return_value).session = mock_sess
            b1 = Budget(name='my_budg')
            b2 = Budget(name='my_budg2')
            t = Transaction(
                budget_amounts={
                    b1: Decimal('10.00'),
                    b2: Decimal('90.00')
                }
            )
            assert len(t.budget_transactions) == 2
            for bt in t.budget_transactions:
                assert isinstance(bt, BudgetTransaction)
                assert bt.transaction == t
            assert {
                bt.budget: bt.amount for bt in t.budget_transactions
            } == {
                b1: Decimal('10.00'),
                b2: Decimal('90.00')
            }
            mock_sess.reset_mock()
            with pytest.raises(AssertionError):
                t.set_budget_amounts({})
        assert mock_sess.mock_calls == []

    def test_add_one(self):
        mock_sess = Mock()
        with patch('%s.inspect' % pbm) as mock_inspect:
            type(mock_inspect.return_value).session = mock_sess
            b1 = Budget(name='my_budg')
            t = Transaction(
                budget_amounts={b1: Decimal('10.00')}
            )
            assert len(t.budget_transactions) == 1
            assert t.budget_transactions[0].transaction == t
            assert t.budget_transactions[0].budget == b1
            assert t.budget_transactions[0].amount == Decimal('10.00')
        # SQLAlchemy 2.0 requires explicit session.add() for BudgetTransactions
        assert len(mock_sess.mock_calls) == 1
        assert mock_sess.mock_calls[0][0] == 'add'

    def test_add_three(self):
        mock_sess = Mock()
        with patch('%s.inspect' % pbm) as mock_inspect:
            type(mock_inspect.return_value).session = mock_sess
            b1 = Budget(name='my_budg')
            b2 = Budget(name='my_budg2')
            b3 = Budget(name='my_budg3')
            t = Transaction(
                budget_amounts={
                    b1: Decimal('50.00'),
                    b2: Decimal('10.00'),
                    b3: Decimal('40.00')
                }
            )
            assert len(t.budget_transactions) == 3
            for bt in t.budget_transactions:
                assert isinstance(bt, BudgetTransaction)
                assert bt.transaction == t
            assert {
                bt.budget: bt.amount for bt in t.budget_transactions
            } == {
                b1: Decimal('50.00'),
                b2: Decimal('10.00'),
                b3: Decimal('40.00')
            }
        # SQLAlchemy 2.0 requires explicit session.add() for BudgetTransactions
        assert len(mock_sess.mock_calls) == 3
        for call in mock_sess.mock_calls:
            assert call[0] == 'add'

    def test_sync(self):
        mock_sess = Mock()
        with patch('%s.inspect' % pbm) as mock_inspect:
            type(mock_inspect.return_value).session = mock_sess
            b1 = Budget(name='my_budg')
            b2 = Budget(name='my_budg2')
            b3 = Budget(name='my_budg3')
            t = Transaction(
                budget_amounts={
                    b1: Decimal('10.00'),
                    b2: Decimal('90.00')
                }
            )
            mock_sess.reset_mock()
            t.set_budget_amounts({
                b1: Decimal('40.00'),
                b3: Decimal('60.00')
            })

        assert len(t.budget_transactions) == 3
        for bt in t.budget_transactions:
            assert isinstance(bt, BudgetTransaction)
            assert bt.transaction == t
        assert {
            bt.budget: bt.amount for bt in t.budget_transactions
        } == {
            b1: Decimal('40.00'),
            b2: Decimal('90.00'),
            b3: Decimal('60.00')
        }

        # SQLAlchemy 2.0 requires explicit session.add() for new BudgetTransactions
        # Expect 1 delete (b2) + 1 add (b3)
        assert len(mock_sess.mock_calls) == 2
        call_names = [c[0] for c in mock_sess.mock_calls]
        assert 'delete' in call_names
        assert 'add' in call_names
        delete_call = [c for c in mock_sess.mock_calls if c[0] == 'delete'][0]
        assert delete_call[1][0].budget == b2
        assert delete_call[1][0].amount == Decimal('90.00')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify, request
from datatables import DataTable
from sqlalchemy import or_
from decimal import Decimal

from biweeklybudget.flaskapp.app import app
from biweeklybudget.db import db_session
from biweeklybudget.models.projects import Project, BoMItem
from biweeklybudget.flaskapp.views.searchableajaxview import SearchableAjaxView
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView


logger = logging.getLogger(__name__)


class ProjectsView(MethodView):
    """
    Render the GET /projects view using the ``projects.html`` template.
    """

    def get(self):
        total_active = Decimal('0.0')
        remain_active = Decimal('0.0')
        for p in db_session.query(Project).filter(
                Project.is_active.__eq__(True)
        ).all():
            total_active += p.total_cost
            remain_active += p.remaining_cost
        return render_template(
            'projects.html',
            total_active=total_active,
            remain_active=remain_active
        )


class ProjectsAjax(SearchableAjaxView):
    """
    Handle GET /ajax/projects endpoint.
    """

    def _filterhack(self, qs, s, args):
        """
        DataTables 1.10.12 has built-in support for filtering based on a value
        in a specific column; when this is done, the filter value is set in
        ``columns[N][search][value]`` where N is the column number. However,
        the python datatables package used here only supports the global
        ``search[value]`` input, not the per-column one.

        However, the DataTable search is implemented by passing a callable
        to ``table.searchable()`` which takes two arguments, the current Query
        that's being built, and the user's ``search[value]`` input; this must
        then return a Query object with the search applied.

        In python datatables 0.4.9, this code path is triggered on
        ``if callable(self.search_func) and search.get("value", None):``

        As such, we can "trick" the table to use per-column searching (currently
        only if global searching is not being used) by examining the per-column
        search values in the request, and setting the search function to one
        (this method) that uses those values instead of the global
        ``search[value]``.

        :param qs: Query currently being built
        :type qs: ``sqlalchemy.orm.query.Query``
        :param s: user search value
        :type s: str
        :param args: args
        :type args: dict
        :return: Query with searching applied
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        # search
        if s != '' and s != 'FILTERHACK':
            if len(s) < 3:
                return qs
            s = '%' + s + '%'
            qs = qs.filter(or_(
                Project.notes.like(s),
                Project.name.like(s)
            ))
        return qs

    def get(self):
        """
        Render and return JSON response for GET /ajax/projects
        """
        args = request.args.to_dict()
        args_dict = self._args_dict(args)
        if self._have_column_search(args_dict) and args['search[value]'] == '':
            args['search[value]'] = 'FILTERHACK'
        table = DataTable(
            args,
            Project,
            db_session.query(Project),
            [
                'name',
                'total_cost',
                'remaining_cost',
                'is_active',
                'notes'
            ]
        )
        table.add_data(
            id=lambda o: o.id
        )
        if args['search[value]'] != '':
            table.searchable(lambda qs, s: self._filterhack(qs, s, args_dict))
        return jsonify(table.json())


class ProjectsFormHandler(FormHandlerView):
    """
    Handle POST /forms/projects
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}
        action = data.get('action', None)
        if action is None:
            raise RuntimeError('No action specified in form data!')
        elif action == 'add':
            if data.get('name', '').strip() == '':
                errors['name'].append('Name cannot be empty')
                have_errors = True
            if len(data.get('name').strip()) > 40:
                errors['name'].append('Name must be <= 40 characters in length')
                have_errors = True
            q = db_session.query(Project).filter(
                Project.name.__eq__(data['name'].strip())
            ).all()
            if len(q) > 0:
                errors['name'].append('Name must be unique')
                have_errors = True
        elif action in ['activate', 'deactivate']:
            if 'id' not in data:
                raise RuntimeError('id must be specified to (de)activate')
            p = db_session.query(Project).get(int(data['id']))
            if p is None:
                raise RuntimeError('Invalid Project id: %s' % data['id'])
        else:
            raise RuntimeError('Invalid action: %s' % action)
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        action = data.get('action', None)
        if action == 'add':
            proj = Project()
            proj.name = data['name'].strip()
            proj.notes = data['notes'].strip()
        elif action == 'activate':
            proj = db_session.query(Project).get(int(data['id']))
            logger.info('Activate %s', proj)
            proj.is_active = True
        elif action == 'deactivate':
            proj = db_session.query(Project).get(int(data['id']))
            logger.info('Deactivate %s', proj)
            proj.is_active = False
        else:
            raise RuntimeError('Invalid action: %s' % action)
        db_session.add(proj)
        db_session.commit()
        if action == 'add':
            logger.info('Created Project %s', proj)
        return 'Successfully saved Project %d in database.' % proj.id


class BoMItemView(MethodView):
    """
    Render the GET /project/<int:project_id> view using the ``bomitem.html``
    template.
    """

    def get(self, project_id):
        proj = db_session.query(Project).get(project_id)
        return render_template(
            'bomitem.html',
            project_id=project_id,
            project_name=proj.name,
            project_notes=proj.notes,
            remaining=proj.remaining_cost,
            total=proj.total_cost
        )


class BoMItemsAjax(SearchableAjaxView):
    """
    Handle GET /ajax/projects/<int:project_id>/bom_items endpoint.
    """

    def _filterhack(self, qs, s, args):
        """
        DataTables 1.10.12 has built-in support for filtering based on a value
        in a specific column; when this is done, the filter value is set in
        ``columns[N][search][value]`` where N is the column number. However,
        the python datatables package used here only supports the global
        ``search[value]`` input, not the per-column one.

        However, the DataTable search is implemented by passing a callable
        to ``table.searchable()`` which takes two arguments, the current Query
        that's being built, and the user's ``search[value]`` input; this must
        then return a Query object with the search applied.

        In python datatables 0.4.9, this code path is triggered on
        ``if callable(self.search_func) and search.get("value", None):``

        As such, we can "trick" the table to use per-column searching (currently
        only if global searching is not being used) by examining the per-column
        search values in the request, and setting the search function to one
        (this method) that uses those values instead of the global
        ``search[value]``.

        :param qs: Query currently being built
        :type qs: ``sqlalchemy.orm.query.Query``
        :param s: user search value
        :type s: str
        :param args: args
        :type args: dict
        :return: Query with searching applied
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        # search
        if s != '' and s != 'FILTERHACK':
            if len(s) < 3:
                return qs
            s = '%' + s + '%'
            qs = qs.filter(or_(
                BoMItem.notes.like(s),
                BoMItem.name.like(s),
                BoMItem.url.like(s)
            ))
        return qs

    def get(self, project_id):
        """
        Render and return JSON response for
        GET /ajax/projects/<int:project_id>/bom_items
        """
        args = request.args.to_dict()
        args_dict = self._args_dict(args)
        if self._have_column_search(args_dict) and args['search[value]'] == '':
            args['search[value]'] = 'FILTERHACK'
        table = DataTable(
            args,
            BoMItem,
            db_session.query(BoMItem).filter(
                BoMItem.project_id.__eq__(project_id)
            ),
            [
                'name',
                'quantity',
                'unit_cost',
                'is_active',
                'notes',
                'url',
                'id'
            ]
        )
        table.add_data(
            line_cost=lambda o: float(o.unit_cost) * o.quantity
        )
        if args['search[value]'] != '':
            table.searchable(lambda qs, s: self._filterhack(qs, s, args_dict))
        return jsonify(table.json())


class ProjectAjax(MethodView):
    """
    Render the GET /ajax/projects/<int:project_id> JSON view.
    """

    def get(self, project_id):
        proj = db_session.query(Project).get(project_id)
        d = proj.as_dict
        d['total_cost'] = proj.total_cost
        d['remaining_cost'] = proj.remaining_cost
        return jsonify(d)


class BoMItemAjax(MethodView):
    """
    Render the GET /ajax/projects/bom_item/<int:id> JSON view.
    """

    def get(self, id):
        return jsonify(db_session.query(BoMItem).get(id).as_dict)


class BoMItemFormHandler(FormHandlerView):
    """
    Handle POST /forms/bom_item
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        errors = {k: [] for k in data.keys()}
        if data.get('name', '').strip() == '':
            errors['name'].append('name cannot be empty')
        errors = self._validate_int('quantity', data, errors)
        errors = self._validate_int('project_id', data, errors)
        if len(errors['quantity']) < 1:
            if int(data['quantity']) < 0:
                errors['quantity'].append('Quantity cannot be negative')
        errors = self._validate_float('unit_cost', data, errors)
        for v in errors.values():
            if len(v) > 0:
                return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        if 'id' in data and data['id'].strip() != '':
            # updating an existing BoMItem
            item = db_session.query(BoMItem).get(int(data['id']))
            if item is None:
                raise RuntimeError("Error: no BoMItem with ID "
                                   "%s" % data['id'])
            action = 'updating BoMItem ' + data['id']
        else:
            item = BoMItem()
            action = 'creating new BoMItem'
        item.project = db_session.query(Project).get(int(data['project_id']))
        item.name = data['name'].strip()
        item.notes = data['notes'].strip()
        item.quantity = int(data['quantity'].strip())
        item.unit_cost = Decimal(data['unit_cost'].strip())
        item.url = data['url'].strip()
        item.is_active = data['is_active']
        logger.info('%s: %s', action, item.as_dict)
        db_session.add(item)
        db_session.commit()
        return {
            'success_message': 'Successfully saved BoMItem %d '
                               'in database.' % item.id,
            'success': True,
            'id': item.id
        }


app.add_url_rule('/projects', view_func=ProjectsView.as_view('projects'))
app.add_url_rule(
    '/forms/projects',
    view_func=ProjectsFormHandler.as_view('projects_form')
)
app.add_url_rule(
    '/ajax/projects',
    view_func=ProjectsAjax.as_view('projects_ajax')
)
app.add_url_rule(
    '/projects/<int:project_id>',
    view_func=BoMItemView.as_view('bom_item_view')
)
app.add_url_rule(
    '/ajax/projects/<int:project_id>',
    view_func=ProjectAjax.as_view('ajax_project_view')
)
app.add_url_rule(
    '/ajax/projects/<int:project_id>/bom_items',
    view_func=BoMItemsAjax.as_view('bom_items_ajax')
)
app.add_url_rule(
    '/ajax/projects/bom_item/<int:id>',
    view_func=BoMItemAjax.as_view('bom_item_ajax')
)
app.add_url_rule(
    '/forms/bom_item',
    view_func=BoMItemFormHandler.as_view('bom_item_form')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from flask.views import MethodView
from flask import render_template, jsonify
from copy import copy
from sqlalchemy import asc

from biweeklybudget.flaskapp.app import app
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.models.account import Account, AcctType, AccountBalance
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.db import db_session
from biweeklybudget.utils import dtnow


class IndexView(MethodView):
    """
    Render the GET / view using the ``index.html`` template.
    """

    def get(self):
        standing = db_session.query(Budget).filter(
            Budget.is_active.__eq__(True), Budget.is_periodic.__eq__(False)
        ).order_by(Budget.name).all()
        pp = BiweeklyPayPeriod.period_for_date(dtnow(), db_session)
        pp_curr_idx = 1
        pp_next_idx = 2
        pp_following_idx = 3
        periods = [pp]
        x = pp
        # add another
        for i in range(0, 8):
            x = x.next
            periods.append(x)
        # trigger calculation/cache of data before passing on to jinja
        for p in periods:
            p.overall_sums
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        active_budgets = {}
        for b in db_session.query(Budget).all():
            k = b.name
            if b.is_income:
                k = '%s (i)' % b.name
            budgets[b.id] = k
            if b.is_active:
                active_budgets[b.id] = k
        return render_template(
            'index.html',
            bank_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Bank,
                Account.is_active == True).all(),  # noqa
            credit_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Credit,
                Account.is_active == True).all(),  # noqa
            investment_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Investment,
                Account.is_active == True).all(),  # noqa
            standing_budgets=standing,
            periods=periods,
            curr_pp=pp,
            pp_curr_idx=pp_curr_idx,
            pp_next_idx=pp_next_idx,
            pp_following_idx=pp_following_idx,
            accts=accts,
            budgets=budgets,
            active_budgets=active_budgets
        )


class AcctBalanaceChartView(MethodView):
    """
    Handle GET /ajax/chart-data/account-balances endpoint.
    """

    def get(self):
        accounts = {
            x.id: x.name for x in db_session.query(Account).all()
        }
        acct_names = accounts.values()
        datedict = {x: None for x in acct_names}
        data = {}
        for bal in db_session.query(AccountBalance).order_by(
            asc(AccountBalance.overall_date)
        ).all():
            ds = bal.overall_date.strftime('%Y-%m-%d')
            if ds not in data:
                data[ds] = copy(datedict)
                data[ds]['date'] = ds
            if bal.ledger is None:
                data[ds][bal.account.name] = 0.0
            else:
                data[ds][bal.account.name] = float(bal.ledger)
        resdata = []
        last = None
        for k in sorted(data.keys()):
            if last is None:
                last = data[k]
                continue
            d = copy(data[k])
            for subk in acct_names:
                if d[subk] is None:
                    d[subk] = last[subk]
            last = d
            resdata.append(d)
        res = {
            'data': resdata,
            'keys': sorted(acct_names)
        }
        return jsonify(res)


app.add_url_rule('/', view_func=IndexView.as_view('index_view'))
app.add_url_rule(
    '/ajax/chart-data/account-balances',
    view_func=AcctBalanaceChartView.as_view('acct_balance_chart_view')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from selenium.webdriver.common.by import By


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestPlaidUpdateView(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/plaid-update')

    def test_1_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Plaid Update - BiweeklyBudget'

    def test_2_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_3_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'

    def test_4_table(self, selenium):
        table = selenium.find_element(By.ID, 'table-items-plaid')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'PlaidItem1',
                'Inst1 (None)',
                'Acct1 (foo), Acct2 (foo), Acct4 (foo4)',
                'now',
                'Update / Fix Item',
                'Refresh'
            ],
            [
                'PlaidItem2',
                'Inst2 (None)',
                'Acct3 (foo)',
                'now',
                'Update / Fix Item',
                'Refresh'
            ],
        ]



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import MetaData


convention = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    # "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}

Base = declarative_base(metadata=MetaData(naming_convention=convention))


class ModelAsDict(object):

    #: Class properties to include in :py:attr:`~.as_dict` result.
    _dict_properties = []

    @property
    def as_dict(self):
        """
        Return a dict representation of the model.

        :return: model's variables/attributes
        :rtype: dict
        """
        d = {}
        for k, v in vars(self).items():
            if not k.startswith('_'):
                d[k] = v
        for k in self._dict_properties:
            d[k] = getattr(self, k)
        return d
